chunks = {}
local nchunks = 0
function get_chunk(X,Y)
   local c = X.."/"..Y
   if not chunks[c] then
      chunks[c] = {}
      chunks[c].types = SC.Construct("PackedArray3D_U8",16,256,16)
      chunks[c].meta = SC.Construct("PackedArray3D_U8",16,256,16)
      chunks[c].light = SC.Construct("PackedArray3D_U8",16,256,16)
      chunks[c].skylight = SC.Construct("PackedArray3D_U8",16,256,16)
      chunks[c].X = X
      chunks[c].Y = Y
      chunks[c].Z = Y
      nchunks = nchunks + 1
   end
   return chunks[c]
end
function kill_chunk(X,Y)
   local c = X.."/"..Y
   chunks[c] = nil
   nchunks = nchunks - 1
end

edits = {}

local screen = assert(SC.Construct("GraphicsDevice", 384, 240, {windowed=true,title="Eye of Thundera"}))
local sw,sh = screen:GetSize()
local hsw,hsh = sw/2,sh/2

local function progress_pos(n)
   local x,y,z
   x = n % 24
   y = math.floor(n / 24) % 15
   z = math.floor(n / 360)
   return x*16,y*16,z
end

local colors = {[-1]={0,0,0},[0]={0,0,1},{1,1,0},{1,0,0},{1,1,1}}
local function place_progress(n)
   local x,y,z = progress_pos(n)
   if colors[z] then
      screen:SetPrimitiveColor(unpack(colors[z]))
      screen:DrawRect(x, y, 16, 16)
   end
end

local function remove_progress(n)
   local x,y,z = progress_pos(n)
   if colors[z] then
      screen:SetPrimitiveColor(unpack(colors[z-1]))
      screen:DrawRect(x, y, 16, 16)
   end
end

local last_progress = 0
local function progress(n)
   if n < 0 then n = 0 end
   if n > last_progress then
      for x=last_progress,n-1 do
	 place_progress(x)
      end
      screen:Update()
   elseif n < last_progress then
      for x=last_progress-1,n,-1 do
	 remove_progress(x)
      end
      screen:Update()
   end
   last_progress = n
end

local projection
do
   local fov = 60*math.pi/180
   local aspect = sw / sh
   local cot = 1/math.tan(fov)
   projection = SC.Construct("Mat4x4", -sw*(cot/aspect), 0, 0, 0, 0, -sh*cot, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0)
end

valid = nil
local last_progress_now
local exempt = {[0]=true,[20]=true,[8]=true,[9]=true,[10]=true,[11]=true}
repeat
   repeat
      coroutine.yield()
   until not not_caring
   if #edits == 0 then
      if last_progress > 0 then
	 progress(0)
      end
      if mode then
	 local mvp
	 do
	    local mat = SC.Construct("Mat4x4",
				     1, 0, 0, 0,
				     0, 1, 0, 0,
				     0, 0, 1, 0,
				     -player_x, -player_stance, -player_z, 1)
	    local yaw = player_yaw * math.pi / 180
	    local c,s = math.cos(yaw),math.sin(yaw)
	    mat = SC.Construct("Mat3x3",
			       c, 0, -s,
			       0, 1, 0,
				  s, 0, c) * mat
	    local pitch = player_pitch * math.pi / 180
	    c,s = math.cos(pitch),math.sin(pitch)
	    mat = SC.Construct("Mat3x3", 
			       1, 0, 0,
			       0, c, -s,
			       0, s, c) * mat
	    mvp = projection * mat
	 end
	 screen:SetPrimitiveColor(0,0,0)
	 screen:DrawRect(0,0,sw,sh)
	 if mode == "beacon" then
	    local beacon = SC.Construct("Vec3", beacon_x, beacon_y, beacon_z)
	    beacon = mvp * beacon
	    if beacon.z > 0.01 then
	       local c = {beacon.x / beacon.w + hsw, beacon.y / beacon.w + hsh}
	       local cc = assert(SCUtil.CompileCoords(c))
	       screen:SetPrimitiveColor(1,1,1)
	       screen:DrawPoints(cc, 10)
	    end
	 elseif mode == "find" then
	    if valid ~= "find" then
	       valid = "find"
	       founds = {}
	       local rem = nchunks
	       for k,v in pairs(chunks) do
		  if rem % 2 == 0 then progress(rem) coroutine.yield() end
		  rem = rem - 1
		  local typ = v.types
		  for x=0,15 do
		     for y=0,255 do
			for z=0,15 do
			   if find_types[1] == 0 then
			      if not exempt[typ:Get(x,y-1,z)] and typ:Get(x,y,z) == 0 and v.light:Get(x,y,z) < 8 then
				 founds[#founds+1] = SC.Construct("Vec3", x+v.X*16, y, z+v.Z*16)
			      end
			   elseif find_types[1] < 0 then
			      if typ:Get(x,y-1,z) == -find_types[1] and typ:Get(x,y,z) == 0 and v.light:Get(x,y,z) < 8 then
				 founds[#founds+1] = SC.Construct("Vec3", x+v.X*16, y, z+v.Z*16)
			      end
			   else
			      if typ:Get(x,y,z) == find_types[1] then
				 founds[#founds+1] = SC.Construct("Vec3", x+v.X*16, y, z+v.Z*16)
			      end
			   end
			end
		     end
		  end
	       end
	       progress(0)
	    end
	    local gotchas = {}
	    for n=1,10 do
	       gotchas[n] = {}
	    end
	    for n=1,#founds do
	       local v = mvp * founds[n]
	       if v.z > 0.01 then
		  local d = math.floor(100/v.w)
		  if d > 10 then d = 10 elseif d < 1 then d = 1 end
		  local gotcha = gotchas[d]
		  gotcha[#gotcha+1] = v.x/v.w+hsw
		  gotcha[#gotcha+1] = v.y/v.w+hsh
	       end
	    end
	    screen:SetPrimitiveColor(1,1,1,0.9)
	    for n=1,10 do
	       local gotcha = gotchas[n]
	       if #gotcha > 0 then
		  local cc = assert(SCUtil.CompileCoords(gotcha))
		  screen:DrawPoints(cc, n)
	       end
	    end
	 end
	 screen:Update()
      end
      coroutine.yield()
   else
      local now = os.time()
      if now ~= last_progress_now then
	 last_progress_now = now
	 progress(#edits)
      end
      local edit = table.remove(edits,1)
      if edit then
	 if edit.type == "one" then
	    local x,y,z = edit.x,edit.y,edit.z
	    local type,dat = edit.typ,edit.dat
	    local X = math.floor(x/16)
	    local Z = math.floor(z/16)
	    local chunk = get_chunk(X,Z)
	    chunk.types:Set(type, x%16,y,z%16)
	    chunk.meta:Set(dat, x%16,y,z%16)
	 elseif edit.type == "multi" then
	    local X,Z,coords,types,datas = edit.X,edit.Z,edit.coords,edit.types,edit.datas
	    local count = edit.count
	    local chunk = get_chunk(X,Z)
	    for n=1,count do
	       local q = coords[n]
	       local x,y,z = math.floor(q/4096),q%256,math.floor(q/256)%16
	       chunk.types:Set(types[n],x,y,z)
	       chunk.meta:Set(datas[n],x,y,z)
	    end
	    coroutine.yield()
	 elseif edit.type == "bulk" then
	    local x,y,z = edit.x,edit.y,edit.z
	    local sizex,sizey,sizez = edit.sizex,edit.sizey,edit.sizez
	    local dat = edit.dat
	    local X = math.floor(x/16)
	    local xo = x%16
	    local yo = y
	    local zo = z%16
	    local Y = math.floor(z/16)
	    local c = get_chunk(X,Y)
	    local spot = 1
	    for x=0,sizex-1 do
	       for z=0,sizez-1 do
		  for y=0,sizey-1 do
		     local type = dat:sub(spot,spot)
		     spot = spot + 1
		     c.types:Set(type:byte(),xo+x,yo+y,zo+z)
		  end
	       end
	    end
	    coroutine.yield()
	    spot = spot + math.floor(sizex*sizey*sizez/2)
	    local spotp = (sizex*sizey*sizez)%2 == 1
	    --[[for x=0,sizex-1 do
	       for z=0,sizez-1 do
		  for y=0,sizey-1 do
		     local meta = dat:sub(spot,spot)
		     meta = meta:byte()
		     if spotp then
			spotp,spot = false,spot + 1
			c.meta:Set(math.floor(meta/16),xo+x,yo+y,zo+z)
		     else
			spotp = true
			c.meta:Set(meta%16,xo+x,yo+y,zo+z)
		     end
		  end
	       end
	    end
	    coroutine.yield()]]
	    for x=0,sizex-1 do
	       for z=0,sizez-1 do
		  for y=0,sizey-1 do
		     local light = dat:sub(spot,spot)
		     light = light:byte()
		     if spotp then
			spotp,spot = false,spot + 1
			c.light:Set(math.floor(light/16),xo+x,yo+y,zo+z)
		     else
			spotp = true
			c.light:Set(light%16,xo+x,yo+y,zo+z)
		     end
		  end
	       end
	    end
	    coroutine.yield()
	    --[[for x=0,sizex-1 do
	       for z=0,sizez-1 do
		  for y=0,sizey-1 do
		     local skylight = dat:sub(spot,spot)
		     skylight = skylight:byte()
		     if spotp then
			spotp,spot = false,spot + 1
			c.skylight:Set(math.floor(skylight/16),xo+x,yo+y,zo+z)
		     else
			spotp = true
			c.skylight:Set(skylight%16,xo+x,yo+y,zo+z)
		     end
		  end
	       end
	    end
	    coroutine.yield()]]
	 elseif edit.type == "kill" then
	    kill_chunk(edit.X, edit.Y)
	 end
      else coroutine.yield() end
   end
until false
