function soft_send(sock, what)
   repeat
      local a,b,c = sock:Send(what)
      local sent = a or c
      if sent == #what then break
      elseif b == "not ready" or not b then
	 if sent then
	    what = what:sub(sent+1,-1)
	 end
	 coroutine.yield("send",sock)
      else
	 error("While doing soft_send to "..sock:GetPrintableAddress()..", got "..tostring(b))
      end
   until false
end

function soft_receive(sock, count, noback)
   if count == 0 then return "" end
   assert(type(count)=="number", "soft_receive only works with number amounts")
   local left = count
   local t = {}
   repeat
      local a,b,c = sock:Receive(math.min(left,65535))
      local got = a or c
      if got and #got > 0 then
	 t[#t+1] = got
	 left = left - #got
	 if other[sock] and not noback then soft_send(other[sock], got) end
	 if left <= 0 then return table.concat(t) end
      end
      if b == "not ready" then
	 coroutine.yield("receive",sock)
      elseif not a then
	 error("While doing soft_receive from "..sock:GetPrintableAddress()..", got "..b)
      end
   until false
end

function maybe_soft_receive(sock, count)
   if count == 0 then return "" end
   assert(type(count)=="number", "maybe_soft_receive only works with number amounts")
   local left = count
   local t = {}
   repeat
      local a,b,c = sock:Receive(math.min(left,65535))
      local got = a or c
      if got and #got > 0 then
	 t[#t+1] = got
	 left = left - #got
	 if other[sock] then soft_send(other[sock], got) else error"Boo" end
	 if left <= 0 then return table.concat(t) end
      end
      if b == "not ready" then
	 if #t == 0 then return nil
	 else
	    coroutine.yield("receive",sock)
	 end
      elseif not a then
	 error("While doing maybe_soft_receive from "..sock:GetPrintableAddress()..", got "..b)
      end
   until false
end

function read8(sock)
   local n = assert(soft_receive(sock, 1))
   local ret = n:byte()
   if ret >= 128 then return ret - 256 else return ret end
end

function read16(sock)
   local n = assert(soft_receive(sock, 2))
   local ret = n:sub(1,1):byte()*256 + n:sub(2,2):byte()
   if ret >= 32768 then return ret - 65536 else return ret end
end

function read32(sock, noback)
   local n = assert(soft_receive(sock, 4, noback))
   local ret = n:sub(1,1):byte()*16777216 + n:sub(2,2):byte()*65536 + n:sub(3,3):byte()*256 + n:sub(4,4):byte()
   if ret >= 2147483648 then return ret - 4294967296 else return ret end
end

function readflt(sock)
   local n = assert(soft_receive(sock, 4))
   local ret = n:sub(2,2):byte()*65536 + n:sub(3,3):byte()*256 + n:sub(4,4):byte()
   if ret < 8388608 then ret = ret + 8388608 end
   ret = ret / 16777216
   local exp = (n:sub(1,1):byte() * 2) % 256 + math.floor(n:sub(2,2):byte()/128)
   exp = exp - 126
   ret = ret * 2^exp
   if n:sub(1,1):byte() >= 128 then ret = -ret end
   return ret
end

function readdbl(sock)
   local n = assert(soft_receive(sock, 8))
   local ret = (n:sub(2,2):byte()%16)*281474976710656 + n:sub(3,3):byte()*1099511627776 + n:sub(4,4):byte()*4294967296 + n:sub(5,5):byte()*16777216 + n:sub(6,6):byte()*65536 + n:sub(7,7):byte()*256 + n:sub(8,8):byte()
   ret = ret / 4503599627370496
   if ret < 1 then ret = ret + 1 end
   local exp = (n:sub(1,1):byte() % 128) * 16 + math.floor(n:sub(2,2):byte() / 16)
   exp = exp - 1023
   ret = ret * 2^exp
   if n:sub(1,1):byte() >= 128 then ret = -ret end
   return ret
end

function readppstring(sock, noback)
   local len = soft_receive(sock, 2, noback)
   if len ~= nil and #len == 2 then
      len = len:sub(1,1):byte()*256 + len:sub(2,2):byte()
      if len == 0 then return "" end
      ppstring = soft_receive(sock, len, noback)
      if ppstring and #ppstring == len then
	 return ppstring
      end
   end
   return nil
end

function make8(n)
   if n < 0 then n = n + 256 end
   return string.char(n)
end

function make16(n)
   if n < 0 then n = n + 65536 end
   return string.char(math.floor(n/256)%256)..string.char(n%256)
end

function make32(n)
   if n < 0 then n = n + 4294967296 end
   return string.char(math.floor(n/16777216))..string.char(math.floor(n/65536)%256)..string.char(math.floor(n/256)%256)..string.char(n%256)
end

function makeppstring(string)
   if #string>32767 then error("ppstring too long!") end
   return string.char(math.floor(#string/256))..string.char(#string%256)..string
end
